"""Kerneloscope helper module."""

import logging
import typing
from urllib import parse

import requests


def _response_hook(
    response: requests.Response, *_: typing.Any, **__: typing.Any
) -> None:
    response.raise_for_status()


class RpcError(Exception):
    """An exception for RPC server error responses."""


class Kerneloscope:
    """A helper class to provide an interface for the kerneloscope RPC API."""

    def __init__(
        self,
        server_url: str,
        *,
        user_agent: str = "kwf-lib.kerneloscope",
        logger: logging.Logger | None = None,
    ):
        """Init."""
        self._server_url = server_url
        self._user_agent = user_agent
        self._logger = logger

    def _log(self, *args: typing.Any, **kwargs: typing.Any) -> None:
        if self._logger:
            self._logger.log(*args, **kwargs)

    def fetch_data(
        self, cmd: str, kerneloscope_args: dict[str, typing.Any] | None = None
    ) -> list[dict[str, typing.Any]]:
        """Return rpc response data."""
        kerneloscope_rpc_url = parse.urljoin(self._server_url, f"/rpc/{cmd}/")
        kerneloscope_args = kerneloscope_args or {}

        with requests.Session() as session:
            session.headers.update({"User-Agent": self._user_agent})
            session.hooks["response"].append(_response_hook)
            response = session.post(kerneloscope_rpc_url, json=kerneloscope_args)

        json_data = response.json()
        self._log(
            logging.DEBUG,
            "Query %s (args: %s) returning: %s",
            cmd,
            kerneloscope_args,
            json_data,
        )

        if "result" in json_data:
            return list(json_data["result"])
        if "error" in json_data:
            self._log(
                logging.INFO, "Query %s returned an error: %s", cmd, json_data["error"]
            )
            raise RpcError(json_data["error"])
        return []

    def get_trees(self) -> list[dict[str, str]]:
        """Return the list of trees known to the kerneloscope instance."""
        return self.fetch_data("get_trees")

    def get_trees_containing(self, commit: str) -> list[dict[str, str]]:
        """Return the list of trees containing the given commit."""
        kerneloscope_args = {"commit": commit}
        return self.fetch_data("get_tree", kerneloscope_args)

    def find_already_backported(self, tree: str, commit: str) -> list[dict[str, str]]:
        """Return a list of commits backported in tree already (if any)."""
        kerneloscope_args = {
            "commit": commit,
            "tree": tree,
            "commit_with_subject": False,
        }
        return self.fetch_data("get_downstream", kerneloscope_args)

    def find_missing_fixes(
        self, tree: str, commit: str
    ) -> tuple[dict[str, typing.Any], dict[str, typing.Any]]:
        """Return the commit hashes w/trees of fixes for the specified commit."""
        kerneloscope_args = {
            "tree": tree,
            "commits": [commit],
            "commit_with_subject": False,
        }
        fixes_data = self.fetch_data("get_missing_fixes", kerneloscope_args)
        legit = {}
        maybe = {}
        for fix in fixes_data:
            if not fix["added"] or fix["merge"]:
                continue
            if fix["kind"]:
                legit[fix["commit"]] = fix["trees"]
            else:
                maybe[fix["commit"]] = fix["trees"]

        if legit or maybe:
            self._log(logging.INFO, "Returning %s and %s", legit, maybe)
        return legit, maybe
